import numpy as np
from .utils import COV_NULL
from .shannon import local_differential_entropy
from numpy.typing import NDArray

def local_total_correlation(
    data: NDArray[np.floating], cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)
) -> NDArray[np.floating]:
    """
    The local total correlation. Note that this measure can be negative.

    .. math:: 
        tc(x) = \\sum_{i=1}^{N}h(x_i) - h(x)

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    inptuts: tuple
        The indices of the channels to include.
    
    Returns
    -------
    NDArray[np.floating]
        The local total correaltion for each frame.

    References
    ----------
    Scagliarini, T., Marinazzo, D., Guo, Y., Stramaglia, S., & Rosas, F. E. (2022).
    Quantifying high-order interdependencies on individual patterns via the local O-information: Theory and applications to music analysis.
    Physical Review Research, 4(1), 013184.
    https://doi.org/10.1103/PhysRevResearch.4.013184

    Pope, M., Varley, T. F., Grazia Puxeddu, M., Faskowitz, J., & Sporns, O. (2025). 
    Time-varying synergy/redundancy dominance in the human cerebral cortex. 
    Journal of Physics: Complexity, 6(1), 015015. 
    https://doi.org/10.1088/2632-072X/adbaa9

    """

    assert cov.shape[0] == data.shape[0], (
        "The data and covariance matrix must have the same dimensionality"
    )

    if idxs[0] == -1:
        _idxs = tuple(i for i in range(data.shape[0]))
    else:
        _idxs = idxs
    N = len(_idxs)

    whole = local_differential_entropy(data[_idxs, :], cov[np.ix_(_idxs, _idxs)])

    sum_parts = np.zeros_like(whole)
    for i in range(N):
        sum_parts += local_differential_entropy(data[_idxs[i], :])

    return sum_parts - whole


def total_correlation(
    cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)
) -> float:
    r"""
    The expected total correlation.

    .. math::

        TC(X) &= D_{KL}(P(X) || \prod_{i=1}^{N}P(X_i) \\
              &= \sum_{i=1}^{N}H(X_i) - H(X)

    For Gaussian random variables, the estimator is:

        .. math::
            \hat{TC}(X) = \frac{-\log R}{2}

    Where :math:`R` is the Pearson correlation matrix.

    Parameters
    ----------
    cov : NDArray[np.floating]
        The covariance matrix.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float
        The expected total correlation.
    
    References
    ----------
    Watanabe, S. (1960). Information Theoretical Analysis of Multivariate Correlation.
    IBM Journal of Research and Development, 4(1), Article 1.
    https://doi.org/10.1147/rd.41.0066
    
    Tononi, G., Sporns, O., & Edelman, G. M. (1994). 
    A measure for brain complexity: Relating functional segregation and integration in the nervous system. 
    Proceedings of the National Academy of Sciences, 91(11), Article 11. 
    https://doi.org/10.1073/pnas.91.11.5033

    Pascual-Marqui, R. D., Kochi, K., & Kinoshita, T. (2025).
    Total/dual correlation/coherence, redundancy/synergy, complexity, and O-information for real and complex valued multivariate data
    (No. arXiv:2507.08773). arXiv.
    https://doi.org/10.48550/arXiv.2507.08773

    """
    if idxs[0] == -1:
        _idxs = tuple(i for i in range(cov.shape[0]))
    else:
        _idxs = idxs

    _cov: NDArray[np.floating] = cov[np.ix_(_idxs, _idxs)]

    # Converting to a correlation/coherence matrix.
    diag: NDArray[np.floating] = np.sqrt(np.diag(_cov))
    d_inv: NDArray[np.floating] = np.diag(1.0 / diag)

    corr: NDArray[np.floating] = d_inv @ _cov @ d_inv

    return -np.linalg.slogdet(corr)[1] / 2


def local_delta_k(
    k: int,
    data: NDArray[np.floating],
    cov: NDArray[np.floating],
    idxs: tuple[int, ...] = (-1,),
) -> NDArray[np.floating]:
    r"""
    A utility function that computes the local generalized form
    of the O-information, S-information, and DTC.

    .. math::
        \delta_{k}(x) = (N-k)tc(x) - \sum_{i=1}^{N} tc(x^{-i})

    Parameters
    ----------
    k : int
        The integer value that defines whether one is computing
        S-info, DTC, or negative O-information.
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    NDArray[np.floating].
        The series of local :math:`\delta^{k}`.

    """
    if idxs[0] == -1:
        _idxs: tuple[int, ...] = tuple(i for i in range(data.shape[0]))
    else:
        _idxs = idxs

    N: int = len(_idxs)
    whole = (N - k) * local_total_correlation(data[_idxs, :], cov[np.ix_(_idxs, _idxs)])

    sum_parts = np.zeros_like(whole)

    for i in range(N):
        idxs = tuple(_idxs[j] for j in range(N) if j != i)
        sum_parts += local_total_correlation(data[idxs, :], cov[np.ix_(idxs, idxs)])

    return whole - sum_parts


def delta_k(k: int, cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)) -> float:
    """
    S-information, DTC, and negative O-information can all be written in a general form:

    .. math::

        WMS^{k}(X) = (N-k)TC(X) - \\sum_{i=1}^{N}TC(X^{-i})

    Parameters
    ----------
    k : int
        The integer value that defines whether one is computing
        S-info, DTC, or negative O-information.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float

    """

    if idxs[0] == -1:
        _idxs: tuple[int, ...] = tuple(i for i in range(cov.shape[0]))
    else:
        _idxs = idxs

    N: int = len(_idxs)

    whole: float = (N - k) * total_correlation(cov[np.ix_(_idxs, _idxs)])
    sum_parts: float = 0.0

    for i in range(N):
        idxs_residual: tuple[int, ...] = tuple(_idxs[j] for j in range(N) if j != i)
        sum_parts += total_correlation(cov[idxs_residual, :][:, idxs_residual])

    return whole - sum_parts


def local_s_information(
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
    idxs: tuple[int, ...] = (-1,),
) -> NDArray[np.floating]:
    """
    Compute local S-information using Gaussian estimation.
    
    .. math::
        \\sigma(X) &= \\sum_{i=1}^{N}i(x_i;x^{-i}) \\\\
                   &= N\\times tc(x) - \\sum_{i=1}^{N}tc(x^{-i}) \\\\
                   &= tc(x) + dtc(x)

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    NDArray[np.floating].
        The series of local S-information.

    """

    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)

    return local_delta_k(k=0, data=data, cov=cov, idxs=idxs)


def s_information(cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)) -> float:
    """
    Compute S-information using Gaussian estimation.

    .. math:: 

        \\Sigma(X) &= \\sum_{i=1}^{N}I(X_i;X^{-i}) \\\\
                   &= N\\times TC(X) - \\sum_{i=1}^{N}TC(X^{-i}) \\\\
                   &= TC(X) + DTC(X)
    
    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns:
    -------
    float
        The expected S-information.

    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate
    Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    return delta_k(k=0, cov=cov, idxs=idxs)


def local_dual_total_correlation(
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
    idxs: tuple[int, ...] = (-1,),
) -> NDArray[np.floating]:
    """
    Computes the dual total correlation using Gaussian estimation. Note that this measure can be negative.
    
    .. math:: 

        dtc(x) &= h(x) - \\sum_{i=1}^{N}h(x_i|x^{-i}) \\\\
               &= (N-1)\\times tc(x) - \\sum_{i=1}^{N}tc(x^{-i})

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    NDArray[np.floating].
        The series of local dual total correlations.

    """

    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)

    return local_delta_k(k=1, data=data, cov=cov, idxs=idxs)


def dual_total_correlation(
    cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)
) -> float:
    """
    Computes the dual total correlation using Gaussian estimation.

    .. math:: 

        DTC(X) &= H(X) - \\sum_{i=1}^{N}H(X_i|X^{-i}) \\\\
               &= (N-1)\\times TC(X) - \\sum_{i=1}^{N}TC(X^{-i})
    
    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution. .
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float
        The expected dual total correlation.

    References
    ----------
    Abdallah, S. A., & Plumbley, M. D. (2012).
    A measure of statistical complexity based on predictive information with application to finite spin systems.
    Physics Letters A, 376(4), 275–281.
    https://doi.org/10.1016/j.physleta.2011.10.066

    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    """

    return delta_k(k=1, cov=cov, idxs=idxs)


def local_o_information(
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
    idxs: tuple[int, ...] = (-1,),
) -> NDArray[np.floating]:
    """
    Computes the local O-information for each sample using Gaussian estimation.

    .. math::

        \\omega(x) &= (2-N)tc(x) + \\sum_{i=1}^{N}tc(x^{-i}) \\\\
                   &= tc(x) - dtc(x)
    
    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    NDArray[np.floating].
        The series of local O-informations.

    References
    ----------
    Scagliarini, T., Marinazzo, D., Guo, Y., Stramaglia, S., & Rosas, F. E. (2022).
    Quantifying high-order interdependencies on individual patterns via the local O-information: Theory and applications to music analysis.
    Physical Review Research, 4(1), 013184.
    https://doi.org/10.1103/PhysRevResearch.4.013184

    Pope, M., Varley, T. F., Grazia Puxeddu, M., Faskowitz, J., & Sporns, O. (2025).
    Time-varying synergy/redundancy dominance in the human cerebral cortex. Journal of Physics: Complexity, 6(1), 015015.
    https://doi.org/10.1088/2632-072X/adbaa9

    """

    return -local_delta_k(k=2, data=data, cov=cov, idxs=idxs)


def o_information(cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)) -> float:
    """
    Compute O-information using Gaussian estimation.
    O-information quantifies the balance between redundancy (positive values) and synergy (negative values) in multivariate information.
    
    .. math::

        \\Omega(X) &= (2-N)TC(X) + \\sum_{i=1}^{N}TC(X^{-i}) \\\\
                   &= TC(X) - DTC(X)

    Parameters
    ----------
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float
        The expected O-information.

    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    return -delta_k(k=2, cov=cov, idxs=idxs)


def tse_complexity(num_samples: int, cov: NDArray[np.floating]) -> float:
    """
    Computes the Tononi-Sporns-Edelman complexity using Gaussian
    estimators.
    
    .. math::

        TSE(X) &= \\sum_{k=1}^{\\lfloor N/2\\rfloor} \\bigg\\langle I(X^{k}_j;X^{-k}_j) \\bigg\\rangle_{j} \\\\
               &= \\sum_{k=2}^{N}\\bigg[\\bigg(\\frac{k}{N}\\bigg)TC(X) - \\langle TC(X^{k}_{j}) \\rangle_{j}  \\bigg] 

    Runtimes scale very badly with system size (as it requires brute-forcing) all possible bipartitions of the system. If the system is too large, a sub-sampling approach is taken: at each scale, num_samples are drawn from the space of bipartitions.

    Parameters
    ----------
    num_samples : int
        The number of sample subsets to compute.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.

    Returns
    -------
    float
        The TSE complexity.

    References
    ----------
    Tononi, G., Sporns, O., & Edelman, G. M. (1994).
    A measure for brain complexity: Relating functional segregation and integration in the nervous system.
    Proceedings of the National Academy of Sciences, 91(11), Article 11.
    https://doi.org/10.1073/pnas.91.11.5033

    """

    N: int = cov.shape[0]  # Number of channels

    tc_whole: float = total_correlation(cov)  # Global total correlation

    null_tcs: NDArray[np.floating] = np.array(
        [(float(i) / float(N)) * tc_whole for i in range(1, N + 1)]
    )
    exp_tcs: NDArray[np.floating] = np.zeros(null_tcs.shape[0])
    exp_tcs[-1] = tc_whole

    for k in range(1, N):
        # All of the samples of subsets at scale i
        samples: NDArray[np.floating] = np.array(
            [np.random.choice(N, size=k, replace=False) for _ in range(num_samples)]
        )
        samples.sort(axis=-1)

        # No need to run the same subset multiple times.
        samples = np.unique(samples, axis=0)

        exp_tcs[k - 1] = np.mean(
            [total_correlation(cov, idxs=tuple(sample)) for sample in samples]
        )

    return (null_tcs - exp_tcs).sum()


def description_complexity(
    cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)
) -> float:
    """
    .. math:: 
        C(X) = \\frac{DTC(X)}{N} 

    Parameters
    ----------
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float
        The expected description complexity.

    References
    ----------
    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w
    
    """

    N: float = float(cov.shape[0]) if idxs[0] == -1 else float(len(idxs))

    return dual_total_correlation(cov=cov, idxs=idxs) / N


def local_description_complexity(
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
    idxs: tuple[int, ...] = (-1,),
) -> NDArray[np.floating]:
    """

    Computes the local description complexity for each sample using Gaussian estimation.

    .. math::
        c(x) = \\frac{dtc(x)}/N

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x time format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If unspecified it is computed directly from the data.
    idxs : tuple, optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    NDArray[np.floating].
        The series of local description complexities.

    """

    N: float = float(cov.shape[0]) if idxs[0] == -1 else float(len(idxs))

    return local_delta_k(k=1, data=data, cov=cov, idxs=idxs) / N



import numpy as np
from numpy.typing import NDArray
from .. import gaussian
from .. import knn


def shannon_entropy(
    discrete_vars: NDArray[np.integer],
    continuous_vars: NDArray[np.floating],
    continuous_estimator: str = "gaussian",
    k: int = 5,
) -> tuple[NDArray[np.floating], float]:
    """
    For discrete :math:`X` and continuous :math:`Y`, leverages the identity

    .. math::
        H(X,Y) = H(X) + H(Y | X)

    Parameters
    ----------
    discrete_vars : NDArray[np.integer]
        Numpy array of the discrete variables, of shape (n_variables, n_samples)
    continuous_vars : NDArray[np.floating]
        Numpy array of the continuous variables, of shape (n_variables, n_samples)
    continuous_estimator : str
        Whether to use a Gaussian or KNN-based estimator of the continuous entropy. Options are "gaussian" or "knn".
    k : int (optional)
        If a KNN-based estimator is selected, sets the k-value. Defaults to 5.

    Returns
    -------
    NDArray[np.floating]
        The local entropy for each sample.
    floating
        The expected entropy over all samples.

    """
    assert discrete_vars.shape[1] == continuous_vars.shape[1], (
        "The discrete and continuous variables must have the same number of samples."
    )
    assert continuous_estimator in {
        "gaussian",
        "knn",
    }, "Continuous estimator options are 'gaussian' or 'knn'."

    unq: NDArray[np.integer]
    counts: NDArray[np.integer]
    unq, counts = np.unique(discrete_vars, axis=1, return_counts=True)

    probs: NDArray[np.floating] = counts / counts.sum()

    avg: float = -np.sum(probs * np.log(probs))
    ptw: NDArray[np.floating] = np.zeros((1, continuous_vars.shape[1]))

    conditional_entropy: float = 0.0
    for i in range(unq.shape[1]):
        state: NDArray[np.integer] = unq[:, [i]]
        mask: NDArray[np.bool_] = np.all(discrete_vars == state, axis=0)

        ptw[:, mask] -= np.log(probs[i])
        continuous_conditional: NDArray[np.floating] = continuous_vars[:, mask]

        if continuous_estimator == "gaussian":
            cov_conditional: NDArray[np.floating] = np.cov(
                continuous_conditional, ddof=0
            )
            ptw_conditional: NDArray[np.floating] = gaussian.local_differential_entropy(
                continuous_conditional, cov_conditional
            )

            ptw[:, mask] += ptw_conditional
            conditional_entropy += probs[i] * ptw_conditional.mean()

        elif continuous_estimator == "knn":
            ptw_conditional, avg_conditional = knn.differential_entropy(
                continuous_conditional, k=k
            )
            ptw[:, mask] += ptw_conditional
            conditional_entropy += probs[i] * avg_conditional

    avg += conditional_entropy

    return ptw, avg


def conditional_entropy(
    discrete_vars: NDArray[np.integer],
    continuous_vars: NDArray[np.floating],
    conditional: str = "discrete",
    continuous_estimator: str = "gaussian",
    k: int = 5,
) -> tuple[NDArray[np.floating], float]:
    """
    For discrete :math:`X` and continuous :math:`Y`, leverages the identity

    .. math::
        H(X | Y) = H(X,Y) - H(Y)

    Either the discrete variables or the continuous variables can be conditioning variable (in which case the other is conditioned).

    Parameters
    ----------
    discrete_vars : NDArray[np.integer]
        Numpy array of the discrete variables, of shape (n_variables, n_samples)
    continuous_vars : NDArray[np.floating]
        Numpy array of the continuous variables, of shape (n_variables, n_samples)
    conditional : str
        Wheter to condition the discrete variables on the continuous, or vice versa.
    continuous_estimator : str
        Whether to use a Gaussian or KNN-based estimator of the continuous entropy. Options are "gaussian" or "knn".
    k : int (optional)
        If a KNN-based estimator is selected, sets the k-value. Defaults to 5.

    Returns
    -------
    NDArray[np.floating]
        The local entropy for each sample.
    floating
        The expected entropy over all samples.

    """
    # %%
    assert discrete_vars.shape[1] == continuous_vars.shape[1], (
        "The discrete and continuous variables must have the same number of samples."
    )
    assert conditional in {
        "discrete",
        "continuous",
    }, "Please specify whether the conditioning variable is discrete or continuous."

    ptw_joint: NDArray[np.floating]
    avg_joint: float
    ptw_joint, avg_joint = shannon_entropy(
        discrete_vars=discrete_vars,
        continuous_vars=continuous_vars,
        continuous_estimator=continuous_estimator,
        k=k,
    )
    ptw_marginal: NDArray[np.floating]
    avg_marginal: float

    if conditional == "continuous":
        if continuous_estimator == "gaussian":
            cov: NDArray[np.floating] = np.cov(continuous_vars, ddof=0)
            ptw_marginal = gaussian.local_differential_entropy(
                continuous_vars, cov=cov
            )
            avg_marginal = ptw_marginal.mean()
        elif continuous_estimator == "knn": 
            ptw_marginal, avg_marginal = knn.differential_entropy(continuous_vars, k)

    elif conditional == "discrete":
        unq: NDArray[np.integer]
        counts: NDArray[np.integer]
        unq, counts = np.unique(discrete_vars, axis=1, return_counts=True)
        probs: NDArray[np.floating] = counts / counts.sum()

        ptw_marginal = np.zeros((1, discrete_vars.shape[1]))
        avg_marginal = -np.sum(probs * np.log(probs))

        for i in range(unq.shape[1]):
            state: NDArray[np.integer] = unq[:, [i]]
            mask: NDArray[np.bool_] = np.all(discrete_vars == state, axis=0)

            ptw_marginal[:, mask] = -np.log(probs[i])

    avg: float = avg_joint - avg_marginal
    ptw: NDArray[np.floating] = ptw_joint - ptw_marginal
    # %%

    return ptw, avg


def mutual_information(
    discrete_vars: NDArray[np.integer], continuous_vars: NDArray[np.floating]
) -> tuple[NDArray[np.floating], float]:
    """
    Returns the local and average mutual information between two (potentially multivariate) discrete and continuous random variables.

    Parameters
    ----------
    discrete_vars : NDArray[np.integer]
        Numpy array of the discrete variables, of shape (n_variables, n_samples)
    continuous_vars : NDArray[np.floating]
        Numpy array of the continuous variables, of shape (n_variables, n_samples)


    Returns
    -------
    NDArray[np.floating]
        The local entropy for each sample.
    floating
        The expected entropy over all samples.


    """
    assert discrete_vars.shape[1] == continuous_vars.shape[1], (
        "The discrete and continuous variables must have the same number of samples."
    )

    cov: NDArray[np.floating] = np.cov(continuous_vars, ddof=0)

    avg: float = gaussian.differential_entropy(cov)
    ptw: NDArray[np.floating] = gaussian.local_differential_entropy(
        continuous_vars, cov
    )

    unq: NDArray[np.integer]
    counts: NDArray[np.integer]
    unq, counts = np.unique(discrete_vars, axis=1, return_counts=True)
    probs: NDArray[np.floating] = counts / counts.sum()

    conditional_entropy = 0.0
    for i in range(unq.shape[1]):
        state: NDArray[np.integer] = unq[:, [i]]
        mask: NDArray[np.bool_] = np.all(discrete_vars == state, axis=0)

        cov_conditional: NDArray[np.floating] = np.cov(continuous_vars[:, mask], ddof=0)
        conditional_entropy += probs[i] * gaussian.differential_entropy(cov_conditional)

        local_conditional: NDArray[np.floating] = gaussian.local_differential_entropy(
            continuous_vars[:, mask], cov_conditional
        )
        ptw[:, mask] -= local_conditional

    avg -= conditional_entropy

    return ptw, avg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb 16 13:01:12 2025

@author: thosvarley
"""

import numpy as np
import random
import math
from numpy.typing import NDArray
from .multivariate_mi import o_information
from typing import Callable


def neg_o_information(x: tuple[NDArray[np.floating], tuple[int, ...]]) -> float:
    """
    A utility function for computing the negative of the O-information
    for a hill-climbing optimizer.

    Parameters
    ----------
    x : tuple[NDArray[np.floating], set[int]]
        A tuple of the arguments to the O-information function.
        The covariance matrix (NDArray) and indices (tuple of ints)

    Returns
    -------
    float
        The negative O-information.

    """
    return -1 * o_information(*x)


def simulated_annealing(
    cov: NDArray[np.floating],
    function: Callable,
    size: int,
    temperature: float = 1.0,
    cooling_rate: float = 0.999,
    min_temperature: float = 1e-5,
    iters_per_temperature: int = 10,
    convergence_window: int = 100,
    convergence_threshold: float = 1e-6,
    verbose: bool = False,
) -> tuple[set[int], float, NDArray[np.floating]]:
    """
    Implements a simulated annealing algorithm for optimizing
    multivariate information measures (O-info, DTC, etc) from
    a covariance matrix. Modified from;

        Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
        Multivariate information theory uncovers synergistic subsystems
        of the human cerebral cortex.
        Communications Biology, 6(1), Article 1.
        https://doi.org/10.1038/s42003-023-04843-w


    Parameters
    ----------
    cov : NDArray[np.floating]
        The covariance matrix that defines the full distribution.
    function : function
        The objective function. Should be taken from the
        syntropy.gaussian.multivariate_mi module.
    size : int
        The size of the ensemble of elements.
    temperature : float, optional
        The initial temperature to initialize the optimization with.
        The default is 1.0.
    cooling_rate : float, optional
        The rate at which the annealing cools.
        Temperature, T, decreases with T(t+1) = T(t)*cooling_rate
        The default is 0.999.
    min_temperature : float, optional
        The stopping temperature. The default is 1e-3.
    iters_per_temperature : int, optional
        How many times to iterate the annealing algorithim before cooling one step.
        The default is 10.
    convergence_window : int, optional
        How many steps back in time to consider for assessing convergence.
        The default is 100.
    convergence_threshold : float, optional
        The standard deviation of the window below which convergence is said
        to have been achieved.. The default is 1e-6.
    verbose : bool, optional
        Whether to print progress messages during optimization.
        The default is False.

    Returns
    -------
    set[int]:
        The indicies of the optimal set.
    float:
        The value of the optimal set.
    list:
        The time series of values.
    """
    assert min_temperature > 0, "The minimum temperature must be greater than zero."

    if verbose:
        print("Initializing annealing...")

    num_steps: int = (
        math.ceil(np.log(min_temperature / temperature) / np.log(cooling_rate)) + 1
    )
    
    chosen_set: set[int] = set(np.random.choice(cov.shape[0], size=size, replace=False))
    best_set: set[int] = chosen_set.copy()
    available_set: set[int] = {i for i in range(cov.shape[0]) if i not in chosen_set}
    
    value: float = function((cov, tuple(chosen_set)))
    best_value: float = 1*value
    values: NDArray[np.floating] = np.zeros(num_steps)

    convergence: bool = False

    counter: int = 1
    if verbose:
        print("Annealing...")
    while temperature > min_temperature:
        for _ in range(iters_per_temperature):
            # Randomly pick an available element and an chosen element
            swap: tuple[int, int] = (
                random.choice(tuple(chosen_set)),
                random.choice(tuple(available_set)),
            )

            # Swap the elements, respectively
            chosen_set.remove(swap[0])
            available_set.remove(swap[1])

            chosen_set.add(swap[1])
            available_set.add(swap[0])

            new_value: float = function((cov, tuple(chosen_set)))

            diff: float = (new_value - value) / abs(
                value
            )  # I assume this will never *actually* be 0 exactly.

            if new_value > best_value:
                best_value = new_value
                best_set = chosen_set.copy()

            # If the new sets are more optimal than the old ones
            # OR the temperature is high enough to accept a sub-optimal configuraiton
            # update the value.
            if (diff > 0) or random.random() < math.exp(diff / temperature):
                value = new_value
            else:
                chosen_set.remove(swap[1])
                available_set.remove(swap[0])

                chosen_set.add(swap[0])
                available_set.add(swap[1])

        # The saved value is the best value of all iters at a given temperature
        values[counter - 1] = value

        if counter > convergence_window:
            std: float = np.std(values[counter - convergence_window : counter])
            if std < convergence_threshold:
                if value < best_value:
                    if verbose:
                        print("...testing convergence...")
                    chosen_set = best_set.copy()
                    available_set = {
                        i for i in range(cov.shape[0]) if i not in chosen_set
                    }
                    value = best_value
                else:
                    if verbose:
                        print("Convergence achieved!")
                    convergence = True
                    break

        temperature *= cooling_rate
        counter += 1

    if convergence is False and verbose:
        print("Annealing schedule finished")
        print("No convergence achieved")

    return best_set, best_value, values[:counter]


def irreducible_synergy(cov: NDArray[np.floating], inputs: tuple[int, ...]) -> bool:
    """
    Computes whether it is possible to remove an element from a
    synergistic sysem and increase the synergy.

    Parameters
    ----------
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.
    inputs : tuple[int, ...]
        The specific elements to test.

    Returns
    -------
    bool
        Returns True if the synergy is irreducible (i.e. there is no element that, when removed, increaes the synergy).
        Returns False otherwise.

    """

    value: float = o_information(cov, inputs)
    assert value < 0, "This only works for initially synergy-dominated subsets."

    N: int = len(inputs)

    for i in range(N):
        reduced_set: tuple[int, ...] = tuple(inputs[j] for j in range(N) if j != i)
        if o_information(cov, reduced_set) < value:
            return False
    return True



import numpy as np
from numpy.typing import NDArray
from scipy.special import digamma
from scipy.spatial import cKDTree
from .utils import check_idxs, build_tree_and_get_distances, get_counts_from_tree


def differential_entropy(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    r"""
    Computes the differential entropy using the Kozachenko-Leoneko estimator.

    .. math::
        \hat{H}(X) = -\psi(k)+\psi(N) + (1/N)\sum_{i=1}^{N}\log d_i

    Parameters
    ----------
    data : NDArray[np.floating]
        Numpy array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating]
        The local differential entropy for each sample.
    float
        The expected differential entropy over all samples

    References
    ----------
    Delattre, S., & Fournier, N. (2017).
    On the Kozachenko–Leonenko entropy estimator.
    Journal of Statistical Planning and Inference, 185, 69–93.
    https://doi.org/10.1016/j.jspi.2017.01.004

    Kozachenko, L. F., & Leonenko, N. N. (1987).
    Sample Estimate of the Entropy of a~Random Vector.
    Problems of Information Transmission, 23(2), 9.

    """

    idxs: tuple[int, ...] = check_idxs(idxs, data.shape[0])

    d: int = len(idxs)
    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N)

    _, distances, _ = build_tree_and_get_distances(data[idxs, :], k=k)

    ptw: NDArray[np.floating] = np.zeros((1, N))
    ptw[0, :] += -psi_k + psi_N + d * np.log(2 * distances[:, -1])

    return ptw, ptw.mean()


def mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    k: int,
    data: NDArray[np.floating],
    algorithm: int = 1,
) -> tuple[NDArray[np.floating], float]:
    """
    A wrapper function for the two KSG mutual information functions.

    See Also
    --------
    mutual_information_1 : Using the KSG-1 algorithm.
    mutual_information_2 : Using the KSG-2 algorithm.

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        Indices of the x-variable.
    idxs_y : tuple[int, ...]
        Indices of the y-variable.
    k : int
        Number of nearest neighbors
    data : NDArray[np.floating]
        Numpy array of shape (n_variables, n_samples)
    algorithm : int
        Whether to use algorithm 1 or 2.
        Defaults to 1

    Returns
    -------
    NDArray[np.floating]
        The local mutual information for each sample.
    float
        The expected mutual information over all samples

    """

    assert algorithm in {1, 2}, "Algorithm must be 1 or 2."

    if algorithm == 1:
        return mutual_information_1(idxs_x=idxs_x, idxs_y=idxs_y, k=k, data=data)
    else:
        return mutual_information_2(idxs_x=idxs_x, idxs_y=idxs_y, k=k, data=data)


def mutual_information_1(
    idxs_x: tuple[int, ...], idxs_y: tuple[int, ...], k: int, data: NDArray[np.floating]
) -> tuple[NDArray[np.floating], float]:
    r"""
    Computes the Kraskov, Stogbauer, Grassberger estimate of the bivariate mutual information
    using the first algorithm presented in Kraskov et. al., (2004)

    .. math::
        \hat{I}(X;Y) = \psi(k) - \psi(N) -\langle \psi(x+1) + \psi(y+1)\rangle

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        Indices of the x-variable.
    idxs_y : tuple[int, ...]
        Indices of the y-variable.
    k : int
        Number of nearest neighbors
    data : NDArray[np.floating]
        Numpy array of shape (n_variables, n_samples)

    Returns
    -------
    NDArray[np.floating
        The local mutual information for each sample.
    float
        The expected mutual information over all samples

    References
    ----------
    Kraskov, A., Stögbauer, H., & Grassberger, P. (2004).
    Estimating mutual information.
    Physical Review E, 69(6), 066138.
    https://doi.org/10.1103/PhysRevE.69.066138

    """

    idxs_xy: tuple[int, ...] = idxs_x + idxs_y

    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N)

    distances: NDArray[np.floating]
    _, distances, _ = build_tree_and_get_distances(data[idxs_xy, :], k=k)

    ptw: NDArray[np.floating] = np.full((1, N), psi_k + psi_N)
    for idxs in (idxs_x, idxs_y):
        tree, _, _ = build_tree_and_get_distances(data[idxs, :], k=k)
        counts: NDArray[np.integer] = get_counts_from_tree(
            tree, data[idxs, :], distances[:, -1]
        )
        ptw[0, :] -= digamma(counts + 1)

    return ptw, ptw.mean()


def mutual_information_2(
    idxs_x: tuple[int, ...], idxs_y: tuple[int, ...], k: int, data: NDArray[np.floating]
) -> tuple[NDArray[np.floating], float]:
    r"""
    Computes the Kraskov, Stogbauer, Grassberger estimate of the bivariate mutual information
    using the second algorithm presented in Kraskove et. al., (2004).

    .. math::
        \hat{I}(X;Y) = \psi(k) - \frac{1}{k} - \psi(N) - \langle \psi(x) + \ldots + \psi(y) \rangle

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        Indices of the x-variable.
    idxs_y : tuple[int, ...]
        Indices of the y-variable.
    k : int
        Number of nearest neighbors
    data : NDArray[np.floating]
        Numpy array of shape (n_variables, n_samples)

    Returns
    -------
    NDArray[np.floating
        The local mutual information for each sample.
    float
        The expected mutual information over all samples

    References
    ----------
    Kraskov, A., Stögbauer, H., & Grassberger, P. (2004).
    Estimating mutual information.
    Physical Review E, 69(6), 066138.
    https://doi.org/10.1103/PhysRevE.69.066138

    """

    idxs_xy: tuple[int, ...] = idxs_x + idxs_y

    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N)

    _, distances, indices = build_tree_and_get_distances(data[idxs_xy, :], k=k)
    neighbors: NDArray[np.integer] = indices[:, 1:]

    ptw: NDArray[np.floating] = np.full((1, N), psi_k - (1 / k) + psi_N)
    for idxs in (idxs_x, idxs_y):
        data_idx: NDArray[np.floating] = data[idxs, :].T
        eps: NDArray[np.floating] = np.repeat(-np.inf, N)

        for j in range(k):
            norm: NDArray[np.floating] = np.linalg.norm(
                data_idx - data_idx[neighbors[:, j]], ord=np.inf, axis=1
            )
            eps = np.maximum(norm, eps)

        tree, distances, _ = build_tree_and_get_distances(data_idx.T, k=k)
        counts = get_counts_from_tree(tree, data_idx.T, eps, strict=False)
        ptw[0, :] -= digamma(counts)

    return ptw, ptw.mean()


def conditional_mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    idxs_z: tuple[int, ...],
    k: int,
    data: NDArray[np.floating],
) -> tuple[NDArray[np.floating], float]:
    """
    Computes the conditional mutual information I(X;Y|Z) using the KSG algorithm 1
    as described in Frenzel & Pompe (2007).

    The conditional mutual information is estimated using:

    .. math::
        \\hat{I}(X;Y|Z) = \\psi(k) - \\langle \\psi(n_{xz}+1) + \\psi(n_{yz}+1) - \\psi(n_z+1) \\rangle

    where n_{xz}, n_{yz}, and n_z are the counts of neighbors within the epsilon-ball
    in the respective subspaces.

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        Indices of the x-variable.
    idxs_y : tuple[int, ...]
        Indices of the y-variable.
    idxs_z : tuple[int, ...]
        Indices of the conditioning variable z.
    k : int
        Number of nearest neighbors
    data : NDArray[np.floating]
        Numpy array of shape (n_variables, n_samples)

    Returns
    -------
    NDArray[np.floating]
        The local conditional mutual information for each sample.
    float
        The expected conditional mutual information over all samples

    References
    ----------
    Frenzel, S., & Pompe, B. (2007).
    Partial Mutual Information for Coupling Analysis of Multivariate Time Series.
    Physical Review Letters, 99(20), 204101.
    https://doi.org/10.1103/PhysRevLett.99.204101

    Kraskov, A., Stögbauer, H., & Grassberger, P. (2004).
    Estimating mutual information.
    Physical Review E, 69(6), 066138.
    https://doi.org/10.1103/PhysRevE.69.066138

    """

    N: int = data.shape[1]
    psi_k: float = digamma(k)

    idxs_xz: tuple[int, ...] = idxs_x + idxs_z
    idxs_yz: tuple[int, ...] = idxs_y + idxs_z
    idxs_joint: tuple[int, ...] = idxs_x + idxs_y + idxs_z

    ptw: NDArray[np.floating] = np.full((1, N), psi_k)

    distances: NDArray[np.floating]
    _, distances, _ = build_tree_and_get_distances(data=data[idxs_joint, :], k=k)

    counter: int = 0
    for idxs in (idxs_z, idxs_xz, idxs_yz):
        tree, _, _ = build_tree_and_get_distances(data[idxs, :], k=k)
        counts: NDArray[np.integer] = get_counts_from_tree(
            tree, data[idxs, :], distances[:, -1]
        )

        if counter == 0:
            ptw[0, :] += digamma(counts + 1)
        else:
            ptw[0, :] -= digamma(counts + 1)
        counter += 1

    return ptw, ptw.mean()



import numpy as np
from numpy.typing import NDArray
from scipy.special import digamma
from scipy.spatial import cKDTree
from .utils import check_idxs, build_tree_and_get_distances, get_counts_from_tree

def total_correlation(
    data: NDArray[np.floating],
    k: int,
    idxs: tuple[int, ...] = (-1,),
    algorithm: int = 1,
) -> tuple[NDArray[np.floating], float]:
    """
    A wrapper function for the two TC functions.

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)
    algorithm : int
        Whether to use algorithm 1 or 2.
        Defaults to 1

    See Also
    --------
    total_correlation_1 : The TC computed using algorithm 1.
    total_correlation_2 : The TC computed using algorithm 2.

    Returns
    -------
    NDArray[np.floating
        The local total correlation for each sample.
    float
        The expected total correlation over all samples

    """

    assert algorithm in {1, 2}, "Algorithm must be 1 or 2."

    if algorithm == 1:
        return total_correlation_1(data=data, k=k, idxs=idxs)
    else:
        return total_correlation_2(data=data, k=k, idxs=idxs)


def total_correlation_1(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    r"""
    Computes the Kraskov, Stogbauer, Grassberger estimate of the total correlation using the first algorithm presented Kraskov et. al (2004)

    .. math::
        \hat{TC}(X) = \psi(k) - (m-1)\psi(N) -\langle \psi(n_{x_{1}}+1) + \ldots + \psi(n_{x_{N}}+1)\rangle

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating
        The local total correlation for each sample.
    float
        The expected total correlation over all samples

    References
    ----------
    Kraskov, A., Stögbauer, H., & Grassberger, P. (2004).
    Estimating mutual information.
    Physical Review E, 69(6), 066138.
    https://doi.org/10.1103/PhysRevE.69.066138

    Watanabe, S. (1960).
    Information Theoretical Analysis of Multivariate Correlation.
    IBM Journal of Research and Development, 4(1), Article 1.
    https://doi.org/10.1147/rd.41.0066

    """
    idxs: tuple[int, ...] = check_idxs(idxs, data.shape[0])

    m: int = len(idxs)
    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N)

    ptw: NDArray[np.floating] = np.full((1, N), psi_k + (m - 1) * psi_N)

    distances: NDArray[np.floating]
    _, distances, _ = build_tree_and_get_distances(data=data[idxs, :], k=k)

    for idx in idxs:
        tree, _, _ = build_tree_and_get_distances(data[(idx,), :], k=k)
        counts: NDArray[np.integer] = get_counts_from_tree(
            tree, data[(idx,), :], distances[:, -1],
        )

        ptw[0, :] -= digamma(counts + 1)

    return ptw, ptw.mean()


def total_correlation_2(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    r"""
    Computes the Kraskov, Stogbauer, Grassberger estimate of the total correlation using the second algorithm presented in Kraskov et. al., (2004).

    .. math::
        \hat{TC}(X) = \psi(k) - ((m-1)/k) - (m-1)\psi(N) - \langle \psi(n_{x_{1}}) + \ldots + \psi(n_{x_{N}}) \rangle

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating
        The local total correlation for each sample.
    float
        The expected total correlation over all samples

    References
    ----------
    Kraskov, A., Stögbauer, H., & Grassberger, P. (2004).
    Estimating mutual information.
    Physical Review E, 69(6), 066138.
    https://doi.org/10.1103/PhysRevE.69.066138

    Watanabe, S. (1960).
    Information Theoretical Analysis of Multivariate Correlation.
    IBM Journal of Research and Development, 4(1), Article 1.
    https://doi.org/10.1147/rd.41.0066

    """
    idxs = check_idxs(idxs, data.shape[0])

    m: int = len(idxs)
    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N)

    ptw: NDArray[np.floating] = np.full(
        (1, N), psi_k - ((m - 1) / k) + ((m - 1) * psi_N)
    )

    distances: NDArray[np.floating]
    indices: NDArray[np.integer]
    _, distances, indices = build_tree_and_get_distances(data[idxs, :], k=k)
    neighbors: NDArray[np.floating] = indices[:, 1:]

    for idx in idxs:
        data_idx: NDArray[np.floating] = data[
            (idx,),
            :,
        ].T

        eps: NDArray[np.floating] = np.full(N, -np.inf)

        for j in range(k):
            norm: NDArray[np.floating] = np.linalg.norm(
                data_idx - data_idx[neighbors[:, j]], ord=np.inf, axis=1
            )
            eps = np.maximum(norm, eps)

        tree, distances, _ = build_tree_and_get_distances(data_idx.T, k=k)
        counts: NDArray[np.integer] = get_counts_from_tree(
            tree, data_idx.T, eps, strict=False
        )
        ptw[0, :] -= digamma(counts)

    return ptw, ptw.mean()


def dual_total_correlation(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    """
    Compute dual total correlation using KSG estimation.
    Code adapted from JIDT
    https://github.com/jlizier/jidt/blob/master/java/source/infodynamics/measures/continuous/kraskov/DualTotalCorrelationCalculatorKraskov.java

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating
        The local dual total correlation for each sample.
    float
        The expected dual total correlation over all samples

    References
    ----------
    Abdallah, S. A., & Plumbley, M. D. (2012).
    A measure of statistical complexity based on predictive information with application to finite spin systems.
    Physics Letters A, 376(4), 275–281.
    https://doi.org/10.1016/j.physleta.2011.10.066

    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    """
    idxs = check_idxs(idxs, data.shape[0])

    m: int = len(idxs)
    N: int = data.shape[1]

    psi_k: float = digamma(k)
    psi_N: float = digamma(N) 

    ptw: NDArray[np.floating] = np.full((1,N), psi_k - psi_N)
    
    distances: NDArray[np.floating]
    tree, distances, _ = build_tree_and_get_distances(data[idxs, :], k=k)
    eps: NDArray[np.floating] = distances[:,-1]

    for i in range(m):

        residual_idxs: list[int] = [idxs[j] for j in range(m) if j != i] 
        residual_data: NDArray[np.floating] = data[residual_idxs,:]

        residual_tree = cKDTree(residual_data.T)

        counts: NDArray[np.integer] = get_counts_from_tree(residual_tree, residual_data, eps)

        ptw[0, :] -= (digamma(counts + 1) - psi_N) / (m-1)
    ptw *= m-1

    return ptw, ptw.mean()


def s_information(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    """
    Compute S-information using KSG estimation.
    Code adapted from JIDT
    https://github.com/jlizier/jidt/blob/master/java/source/infodynamics/measures/continuous/kraskov/SInfoCalculatorKraskov.java

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating
        The local S-information for each sample.
    float
        The expected S-information over all samples

    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """
    idxs_: tuple[int, ...] = check_idxs(idxs, data.shape[0])

    N: int = data.shape[1]
    m: int = len(idxs_)

    psi_k, psi_N = digamma([k, N])

    # Build tree for joint distribution (all m dimensions)
    tree, distances, _ = build_tree_and_get_distances(data[idxs_, :], k=k)
    eps: NDArray[np.floating] = distances[:, -1]

    # Initialize local values: start with (ψ(k) - ψ(N))
    ptw: NDArray[np.floating] = np.full((1,N), psi_k - psi_N)

    # For each dimension d
    for d in range(m):
        # Small marginal: just dimension d alone (1D)
        small_marginal_data = data[idxs_[d] : idxs_[d] + 1, :].T  # Shape: (N, 1)
        tree_small = cKDTree(small_marginal_data)

        # Big marginal: all dimensions except d (m-1 dimensions)
        big_marginal_idxs = [idxs_[j] for j in range(m) if j != d]
        big_marginal_data = data[big_marginal_idxs, :].T  # Shape: (N, m-1)
        tree_big = cKDTree(big_marginal_data)

        counts_small = get_counts_from_tree(tree_small, small_marginal_data.T, eps)
        counts_big = get_counts_from_tree(tree_big, big_marginal_data.T, eps)

        # Subtract contributions from both marginals, divided by m
        ptw[0,:] -= (digamma(counts_big + 1) - psi_N) / m
        ptw[0,:] -= (digamma(counts_small + 1) - psi_N) / m

    # Multiply everything by m
    ptw *= m

    return ptw, ptw.mean()


def o_information(
    data: NDArray[np.floating], k: int, idxs: tuple[int, ...] = (-1,)
) -> tuple[NDArray[np.floating], float]:
    """
    Compute O-information using KSG estimation.
    Code adapted from JIDT
    https://github.com/jlizier/jidt/blob/master/java/source/infodynamics/measures/continuous/kraskov/OInfoCalculatorKraskov.java

    O-information quantifies the balance between redundancy (positive values)
    and synergy (negative values) in multivariate information.

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)

    Returns
    -------
    NDArray[np.floating
        The local O-information for each sample.
    float
        The expected O-information over all samples

    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate
    Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """
    if idxs[0] == -1:
        idxs_ = tuple(range(data.shape[0]))
    else:
        idxs_ = idxs

    N: int = data.shape[1]
    m: int = len(idxs_)

    # Special case: O-information is 0 for 2D data
    if m == 2:
        return np.zeros(N), 0.0

    psi_k, psi_N = digamma([k, N])

    tree, distances, _ = build_tree_and_get_distances(data[idxs_, :], k=k)
    eps: NDArray[np.floating] = distances[:, -1]

    # Initialize local values: start with (ψ(k) - ψ(N))
    ptw: NDArray[np.floating] = np.full((1,N), psi_k - psi_N)

    for d in range(m):
        # Small marginal: just dimension d alone (1D)
        small_marginal_data = data[idxs_[d] : idxs_[d] + 1, :]  # Shape: (N, 1)
        tree_small = cKDTree(small_marginal_data.T)

        # Big marginal: all dimensions except d (m-1 dimensions)
        big_marginal_idxs = [idxs_[j] for j in range(m) if j != d]
        big_marginal_data = data[big_marginal_idxs, :]  # Shape: (N, m-1)
        tree_big = cKDTree(big_marginal_data.T)

        counts_small = get_counts_from_tree(tree_small, small_marginal_data, eps)
        counts_big = get_counts_from_tree(tree_big, big_marginal_data, eps)

        ptw[0,:] -= (digamma(counts_big + 1) - psi_N) / (m - 2)
        ptw[0,:] += (digamma(counts_small + 1) - psi_N) / (m - 2)

    # Multiply everything by (2 - m)
    ptw *= 2 - m

    return ptw, ptw.mean()



import numpy as np
import itertools as it
from numpy.typing import NDArray


COV_NULL: NDArray[np.floating] = np.array([[-1.0]])

# %% LIBRARY


def check_cov(
    cov: NDArray[np.floating], data: NDArray[np.floating]
) -> NDArray[np.floating]:

    if cov.shape == ():
        cov_ = cov 
    else:
        if cov[0, 0] == -1:
            cov_ = np.cov(data, ddof=0.0)
        else:
            cov_ = cov.copy()

    return cov_


def make_powerset(iterable):
    """
    Computes the powerset of a collection of elements.

    .. math:: 
        \\mathcal{P}(\\{X_1,X_2,X_3\\}) \\to (\\{\\}, \\{X_1\\}, \\{X_2\\}, \\{X_3\\}, \\{X_1,X_2\\}, \\{X_1,X_3\\}, \\{X_1,X_2,X_3\\} )

    """
    xs: list = list(iterable)
    # note we return an iterator rather than a list
    return it.chain.from_iterable(it.combinations(xs, n) for n in range(len(xs) + 1))




def correlation_to_mutual_information(cov: NDArray[np.floating]) -> NDArray[np.floating]:
    """
    Converts a Pearson correlation matrix to a Guassian mutual
    information matrix based on the identity:

    :math:`I(X;Y) = \\frac{-\\log(1-(r_{XY})^2)}{2}`

    Where :math:`r_{XY}` is the Pearson correlation coefficient between :math:`X` and :math:`Y`.

    Also works for a covariance matrix if the processes have 0 mean
    and unit variance.

    Parameters
    ----------
    cov : NDArray[np.floating]
        A covariance matrix.

    Returns
    -------
    NDArray[np.floating]
        The equivalent mutual information matrix.

    """
    mi = -np.log(1 - (cov**2)) / 2.0
    np.fill_diagonal(mi, np.nan)

    return mi



import numpy as np
import scipy.signal as signal
import scipy.integrate as integrate
from numpy.typing import NDArray


def construct_csd_tensor(
    idxs: tuple[int, ...], data: NDArray[np.floating], fs: int = 1, nperseg: int = 1024
) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
    """

    Parameters
    ----------
    idxs : tuple[int, ...]

    data : NDArray[np.floating]

    fs : int

    nperseg : int


    Returns
    -------
    NDArray[np.floating]


    """
    N: int = len(idxs)
    S: NDArray[np.floating] = np.zeros((nperseg, N, N))

    for i in range(N):
        for j in range(i + 1):
            f, Pij = signal.csd(
                data[idxs[i], :],
                data[idxs[j], :],
                return_onesided=False,
                fs=fs,
                nperseg=nperseg,
            )
            Pij = np.fft.fftshift(Pij)
            S[:, i, j] = Pij.real
            S[:, j, i] = np.conj(Pij).real

    omega = 2 * np.pi * f / fs
    omega = np.fft.fftshift(omega)

    return S, omega


def differential_entropy_rate(
    idxs: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
) -> tuple[NDArray[np.floating], float]:
    """
    Computes the differential entropy rate of a potentially multivariate stochastic process.

    .. math::
        H(X)=\\frac{1}{4\\pi} \int_{-\\pi}^{\\pi} \\log \\left( (2\\pi e)^N |S_X(\\omega)| \\right) \\, d\\omega

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the analysis.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local entropies for each frequency band.
    float
        The average entropy across the whole spectrum.
    """
    # %%
    N = len(idxs)

    S, omega = construct_csd_tensor(idxs=idxs, data=data, fs=fs, nperseg=nperseg)

    ptw = np.array(
        [(1 / 2) * np.log(((2 * np.pi * np.e) ** N) * np.linalg.det(a)) for a in S]
    )
    avg = (1 / (2 * np.pi)) * integrate.simpson(ptw, omega)
    # %%
    return ptw, avg


def mutual_information_rate(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
) -> tuple[NDArray[np.floating], float]:
    """
    Computes the mutual information rate between two (potentially multivariate) Gaussian processes.

    .. math:: 
        I(X; Y) = \\frac{1}{4\\pi} \\int_{-\\pi}^{\\pi} \\log \\left( \\frac{  |S_X(\\omega)||S_Y(\\omega)| }{ |S_{XY}(\\omega)| } \\right) d\\omega

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        The indices of the channels to include in the inputs.
    idxs_y : tuple[int, ...]
        The indices of the channels to include in the target.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local mutual informations for each frequency band.
    float
        The average mutual information across the whole spectrum.
    
    References
    ----------
    Faes, L., Sparacino, L., Mijatovic, G., Antonacci, Y., Ricci, L., Marinazzo, D., & Stramaglia, S. (2025).
    Partial Information Rate Decomposition (No. arXiv:2502.04550). arXiv.
    https://doi.org/10.48550/arXiv.2502.04550

    Faes, L., Pernice, R., Mijatovic, G., Antonacci, Y., Krohova, J. C., Javorka, M., & Porta, A. (2021).
    Information decomposition in the frequency domain: A new framework to study cardiovascular and cardiorespiratory oscillations.
    Philosophical Transactions of the Royal Society A: Mathematical, Physical and Engineering Sciences, 379(2212), 20200250.
    https://doi.org/10.1098/rsta.2020.0250

    """
    Nx: int = len(idxs_x)
    Ny: int = len(idxs_y)

    idxs_: tuple[int, ...] = idxs_x + idxs_y

    reidxs_x: tuple[int, ...] = tuple(i for i in range(Nx))
    reidxs_y: tuple[int, ...] = tuple(i + Nx for i in range(Ny))

    S, omega = construct_csd_tensor(idxs=idxs_, data=data, fs=fs, nperseg=nperseg)

    log_x: NDArray[np.floating] = np.array(
        [np.linalg.slogdet(a[np.ix_(reidxs_x, reidxs_x)])[1] for a in S]
    )
    log_y: NDArray[np.floating] = np.array(
        [np.linalg.slogdet(a[np.ix_(reidxs_y, reidxs_y)])[1] for a in S]
    )
    log_xy: NDArray[np.floating] = np.array([np.linalg.slogdet(a)[1] for a in S])

    ptw = (1 / 2) * (log_x + log_y - log_xy)
    avg = (1 / (2 * np.pi)) * integrate.simpson(ptw, omega)

    return ptw, avg


def total_correlation_rate(
    idxs: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
) -> tuple[NDArray[np.floating], float]:
    """
    A straightforward extension of the mutual information rate to the total correlation.

    .. math:: 
        TC(X,Y,\\ldots,Z) = \\frac{1}{4\\pi} \\int_{-\\pi}^{\\pi} \\log \\left( \\frac{  |S_X(\\omega)||S_Y(\\omega)|\\ldots|S_Z(\\omega)| }{ |S_{XY\\ldots Z}(\\omega)| } \\right) d\\omega

    WARNING: As far as I know this TC rate idea has never been formally explored before. It should work fine as a natural generalization of the MI, but it hasn't ever been published or peer reviewed.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the total correlation calculation.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local total correlation  for each frequency band.
    float
        The average total correlation across the whole spectrum.
    """

    S, omega = construct_csd_tensor(idxs=idxs, data=data, fs=fs, nperseg=nperseg)

    sum_parts = np.array([np.log(np.diag(a)) for a in S]).sum(axis=-1)

    whole = np.array([np.linalg.slogdet(a)[1] for a in S])

    ptw = (1 / 2) * (sum_parts - whole)
    avg = (1 / (2 * np.pi)) * integrate.simpson(ptw, omega)

    return ptw, avg


def k_wms_rate(
    idxs: tuple[int, ...],
    k: int,
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
    verbose: bool = False,
) -> tuple[NDArray[np.floating], float]:
    """
    A straightforward extension of the total correlation rate to the K whole-minus-sum rate.

    Recall that S-information, DTC, and negative O-information can all be written in a general form:

    .. math:: 
        WMS^{k}(X) = (N-k)TC(X) - \\sum_{i=1}^{N}TC(X^{-i})

    WARNING: As far as I know this rate idea has never been formally explored before. It should work fine as a natural generalization of the MI, but it hasn't even been published or peer reviewed.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the calculation.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local k_wms rate for each frequency band.
    float
        The average k_wms across the whole spectrum.
    
    References
    ----------
    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w
    
    """
    
    N0: int = len(idxs)
    ptw_whole: NDArray[np.floating]
    avg_whole: float

    ptw_whole, avg_whole = total_correlation_rate(
        idxs=idxs, data=data, fs=fs, nperseg=nperseg
    )

    ptw_whole *= (N0 - k)
    avg_whole *= (N0 - k)

    ptw_sum_parts: NDArray[np.floating] = np.zeros_like(ptw_whole)
    avg_sum_parts: float = 0.0

    for i in range(N0):
        idxs_residual: tuple[idxs, ...] = tuple(idxs[j] for j in range(N0) if j != i)

        ptw_residuals, avg_residuals = total_correlation_rate(
            idxs=idxs_residual, data=data, fs=fs, nperseg=nperseg
        )

        ptw_sum_parts += ptw_residuals
        avg_sum_parts += avg_residuals

        if verbose is True:
            print((i + 1) / N0)

    return ptw_whole - ptw_sum_parts, avg_whole - avg_sum_parts


def s_information_rate(
    idxs: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
    verbose: bool = False,
) -> tuple[NDArray[np.floating], float]:
    """
    A straightforward extension of the S-information rate from the total correlation rate.

    The S-information is equivalant to:

    .. math:: 
        \\Sigma(X) = \\sum_{i=1}^{N}I(X_i;X^{-i})

    .. math:: 
        \\Sigma(X) = TC(X) + DTC(X)

    WARNING: As far as I know this rate idea has never been formally explored before. It should work fine as a natural generalization of the MI, but it hasn't even been published or peer reviewed.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the calculation.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local S-information rate for each frequency band.
    float
        The average S-information across the whole spectrum.
    
    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    ptw_s, avg_s = k_wms_rate(
        idxs=idxs, k=0, data=data, fs=fs, nperseg=nperseg, verbose=verbose
    )
    return ptw_s, avg_s


def dual_total_correlation_rate(
    idxs: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
    verbose: bool = False,
) -> tuple[NDArray[np.floating], float]:
    """
    A straightforward extension of the dual total correlation rate from the total correlation rate.

    The dual total correlation is given alternately by:

    .. math:: 
        DTC(X) = H(X) - \\sum_{i=1}^{N}H(X_i|X^{-i})

    .. math:: 
        DTC(X) = (N-1)TC(X) - \\sum_{i=1}^{N}TC(X^{-i})

    WARNING: As far as I know this rate idea has never been formally explored before. It should work fine as a natural generalization of the MI, but it hasn't even been published or peer reviewed.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the calculation.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local dual total correlation rate for each frequency band.
    float
        The average dual total correlation across the whole spectrum.
    
    References
    ----------
    Abdallah, S. A., & Plumbley, M. D. (2012).
    A measure of statistical complexity based on predictive information with application to finite spin systems.
    Physics Letters A, 376(4), 275–281.
    https://doi.org/10.1016/j.physleta.2011.10.066

    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    """

    ptw_dtc, avg_dtc = k_wms_rate(
        idxs=idxs, k=1, data=data, fs=fs, nperseg=nperseg, verbose=verbose
    )
    return ptw_dtc, avg_dtc


def o_information_rate(
    idxs: tuple[int, ...],
    data: NDArray[np.floating],
    fs: int = 1,
    nperseg: int = 1024,
    verbose: bool = False,
) -> tuple[NDArray[np.floating], float]:
    """
    A straightforward extension of the O-information rate from the total correlation rate.

    .. math::   
        \\Omega(X) = (2-N)TC(X) + \\sum_{i=1}^{N}TC(X^{-i}) 

    .. math:: 
        \\Omega(X) = TC(X) - DTC(X)

    WARNING: As far as I know this rate idea has never been formally explored before. It should work fine as a natural generalization of the MI, but it hasn't even been published or peer reviewed.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The indices of the channels to include in the calculation.
    data : NDArray[np.floating]
        The data in channels x samples format.
    fs: int
        The sampling rate of the time series data in Hz.
        The default is 1.
    nperseg: int
        The number of samples to include in each segment.
        Passed to the various scipy.signal functions that compute the spectral analyses.
        The default is 1024

    Returns
    -------
    NDArray[np.floating]
        The local O-information rate for each frequency band.
    float
        The average O-information across the whole spectrum.
    
    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    ptw_o, avg_o = k_wms_rate(
        idxs=idxs, k=2, data=data, fs=fs, nperseg=nperseg, verbose=verbose
    )
    return -ptw_o, -avg_o
